from typing import Callable, Optional

import numpy as np
from numpy.typing import ArrayLike
from scipy.special import legendre, jn
import dask.array as darray
from functools import partial
from scipy.spatial import KDTree

from .autosave import autosave_data
from .utils import coherent_sum
from .pbc import pbc_diff, pbc_points
from .coordinates import Coordinates, CoordinateFrame, displacements_without_drift


def _is_multi_selector(selection):
    if len(selection) == 0:
        return False
    elif (
        isinstance(selection[0], int)
        or isinstance(selection[0], bool)
        or isinstance(selection[0], np.integer)
        or isinstance(selection[0], np.bool_)
    ):
        return False
    else:
        for indices in selection:
            if len(indices) == 0:
                continue
            elif (
                isinstance(indices[0], int)
                or isinstance(indices[0], bool)
                or isinstance(indices[0], np.integer)
                or isinstance(indices[0], np.bool_)
            ):
                return True
            else:
                raise ValueError(
                    "selector has more than two dimensions or does not "
                    "contain int or bool types"
                )


def _calc_correlation(
    frames: Coordinates,
    start_frame: CoordinateFrame,
    function: Callable,
    selection: np.ndarray,
    shifted_idx: np.ndarray,
) -> np.ndarray:
    if len(selection) == 0:
        correlation = np.zeros(len(shifted_idx))
    else:
        start = start_frame[selection]
        correlation = np.array(
            [
                function(start, frames[frame_index][selection])
                for frame_index in shifted_idx
            ]
        )
    return correlation


def _calc_correlation_multi(
    frames: Coordinates,
    start_frame: CoordinateFrame,
    function: Callable,
    selection: np.ndarray,
    shifted_idx: np.ndarray,
) -> np.ndarray:
    correlations = np.zeros((len(selection), len(shifted_idx)))
    for i, frame_index in enumerate(shifted_idx):
        frame = frames[frame_index]
        for j, current_selection in enumerate(selection):
            if len(selection) == 0:
                correlations[j, i] = 0
            else:
                correlations[j, i] = function(
                    start_frame[current_selection], frame[current_selection]
                )
    return correlations


def _average_correlation(result):
    averaged_result = []
    for n in range(result.shape[1]):
        clean_result = []
        for entry in result[:, n]:
            if np.all(entry == 0):
                continue
            else:
                clean_result.append(entry)
        averaged_result.append(np.average(np.array(clean_result), axis=0))
    return np.array(averaged_result)


def _average_correlation_multi(result):
    clean_result = []
    for entry in result:
        if np.all(entry == 0):
            continue
        else:
            clean_result.append(entry)
    return np.average(np.array(clean_result), axis=0)


def shifted_correlation(
    function: Callable,
    frames: Coordinates,
    selector: Optional[Callable] = None,
    segments: int = 10,
    skip: float = 0.1,
    window: float = 0.5,
    average: bool = True,
    points: int = 100,
) -> tuple[np.ndarray, np.ndarray]:
    """Compute a time-dependent correlation function for a given trajectory.

    To improve statistics, multiple (possibly overlapping) windows will be
    layed over the whole trajectory and the correlation is computed for them separately.
    The start frames of the windows are spaced linearly over the valid region of
    the trajectory (skipping frames in the beginning given by skip parameter).

    The points within each window are spaced logarithmically.

    Only a certain subset of the given atoms may be selected for each window
    individually using a selector function.

    Note that this function is specifically optimized for multi selectors, which select
    multiple selection sets per window, for which the correlation is to be computed
    separately.


    Arguments
    ---------
    function:
        The (correlation) function to evaluate.
        Should be of the form (CoordinateFrame, CoordinateFrame) -> float

    frames:
        Trajectory to evaluate on

    selector: (optional)
        Selection function so select only certain selection sets for each start frame.
        Should be of the form
            (CoordinateFrame) -> list[A]
        where A is something you can index an ndarray with.
        For example a list of indices or a bool array.
        Must return the same number of selection sets for every frame.

    segments:
        Number of start frames

    skip:
        Percentage of trajectory to skip from the start

    window:
        Length of each segment given as percentage of trajectory

    average:
        Whether to return averaged results.
        See below for details on the returned ndarray.

    points:
        Number of points per segment


    Returns
    -------
    times: ndarray
        1d array of time differences to start frame
    result: ndarray
        2d ndarray of averaged (or non-averaged) correlations.

        When average==True (default) the returned array will be of the shape (S, P)
        where S is the number of selection sets and P the number of points per window.
        For selection sets that where empty for all start frames all data points will be
        zero.

        When average==False the returned array will be of shape (W, S) with
        dtype=object. The elements are either ndarrays of shape (P,) containing the
        correlation data for the specific window and selection set or None if the
        corresponding selection set was empty.
        W is the number of segments (windows).
        S and P are the same as for average==True.

    """
    if 1 - skip < window:
        window = 1 - skip

    start_frame_indices = np.unique(
        np.linspace(
            len(frames) * skip,
            len(frames) * (1 - window),
            num=segments,
            endpoint=False,
            dtype=int,
        )
    )

    num_frames_per_window = int(len(frames) * window)
    logspaced_indices = np.logspace(0, np.log10(num_frames_per_window + 1), num=points)
    logspaced_indices = np.unique(np.int_(logspaced_indices) - 1)
    logspaced_time = (
        np.array([frames[i].time for i in logspaced_indices]) - frames[0].time
    )

    if selector is None:
        multi_selector = False
    else:
        selection = selector(frames[0])
        multi_selector = _is_multi_selector(selection)

    result = []
    for start_frame_index in start_frame_indices:
        shifted_idx = logspaced_indices + start_frame_index
        start_frame = frames[start_frame_index]
        if selector is None:
            selection = np.arange(len(start_frame))
        else:
            selection = selector(start_frame)
        if multi_selector:
            result_segment = _calc_correlation_multi(
                frames, start_frame, function, selection, shifted_idx
            )
        else:
            result_segment = _calc_correlation(
                frames, start_frame, function, selection, shifted_idx
            )
        result.append(result_segment)
    result = np.array(result)

    if average:
        if multi_selector:
            result = _average_correlation_multi(result)
        else:
            result = _average_correlation(result)
    return logspaced_time, result


def msd(
    start_frame: CoordinateFrame,
    end_frame: CoordinateFrame,
    trajectory: Coordinates = None,
    axis: str = "all",
) -> float:
    """
    Mean square displacement
    """
    if trajectory is None:
        displacements = start_frame - end_frame
    else:
        displacements = displacements_without_drift(start_frame, end_frame, trajectory)
    if axis == "all":
        return (displacements**2).sum(axis=1).mean()
    elif axis == "xy" or axis == "yx":
        return (displacements[:, [0, 1]] ** 2).sum(axis=1).mean()
    elif axis == "xz" or axis == "zx":
        return (displacements[:, [0, 2]] ** 2).sum(axis=1).mean()
    elif axis == "yz" or axis == "zy":
        return (displacements[:, [1, 2]] ** 2).sum(axis=1).mean()
    elif axis == "x":
        return (displacements[:, 0] ** 2).mean()
    elif axis == "y":
        return (displacements[:, 1] ** 2).mean()
    elif axis == "z":
        return (displacements[:, 2] ** 2).mean()
    else:
        raise ValueError('Parameter axis has to be ether "all", "x", "y", or "z"!')


def isf(
    start_frame: CoordinateFrame,
    end_frame: CoordinateFrame,
    q: float = 22.7,
    trajectory: Coordinates = None,
    axis: str = "all",
) -> float:
    """
    Incoherent intermediate scattering function. To specify q, use
    water_isf = functools.partial(isf, q=22.77) # q has the value 22.77 nm^-1
    """
    if trajectory is None:
        displacements = start_frame - end_frame
    else:
        displacements = displacements_without_drift(start_frame, end_frame, trajectory)
    if axis == "all":
        distance = (displacements**2).sum(axis=1) ** 0.5
        return np.sinc(distance * q / np.pi).mean()
    elif axis == "xy" or axis == "yx":
        distance = (displacements[:, [0, 1]] ** 2).sum(axis=1) ** 0.5
        return np.real(jn(0, distance * q)).mean()
    elif axis == "xz" or axis == "zx":
        distance = (displacements[:, [0, 2]] ** 2).sum(axis=1) ** 0.5
        return np.real(jn(0, distance * q)).mean()
    elif axis == "yz" or axis == "zy":
        distance = (displacements[:, [1, 2]] ** 2).sum(axis=1) ** 0.5
        return np.real(jn(0, distance * q)).mean()
    elif axis == "x":
        distance = np.abs(displacements[:, 0])
        return np.mean(np.cos(np.abs(q * distance)))
    elif axis == "y":
        distance = np.abs(displacements[:, 1])
        return np.mean(np.cos(np.abs(q * distance)))
    elif axis == "z":
        distance = np.abs(displacements[:, 2])
        return np.mean(np.cos(np.abs(q * distance)))
    else:
        raise ValueError('Parameter axis has to be ether "all", "x", "y", or "z"!')


def rotational_autocorrelation(
    start_frame: CoordinateFrame, end_frame: CoordinateFrame, order: int = 2
) -> float:
    """
    Compute the rotational autocorrelation of the legendre polynomial for the
    given vectors.

    Args:
        start_frame, end_frame: CoordinateFrames of vectors
        order (opt.): Order of the legendre polynomial.

    Returns:
        Scalar value of the correlation function.
    """
    scalar_prod = (start_frame * end_frame).sum(axis=-1)
    poly = legendre(order)
    return poly(scalar_prod).mean()


def van_hove_self(
    start_frame: CoordinateFrame,
    end_frame: CoordinateFrame,
    bins: ArrayLike,
    trajectory: Coordinates = None,
    axis: str = "all",
) -> np.ndarray:
    r"""
    Compute the self-part of the Van Hove autocorrelation function.

    ..math::
      G(r, t) = \sum_i \delta(|\vec r_i(0) - \vec r_i(t)| - r)
    """
    if trajectory is None:
        vectors = start_frame - end_frame
    else:
        vectors = displacements_without_drift(start_frame, end_frame, trajectory)
    if axis == "all":
        delta_r = (vectors**2).sum(axis=1) ** 0.5
    elif axis == "xy" or axis == "yx":
        delta_r = (vectors[:, [0, 1]] ** 2).sum(axis=1) ** 0.5
    elif axis == "xz" or axis == "zx":
        delta_r = (vectors[:, [0, 2]] ** 2).sum(axis=1) ** 0.5
    elif axis == "yz" or axis == "zy":
        delta_r = (vectors[:, [1, 2]] ** 2).sum(axis=1) ** 0.5
    elif axis == "x":
        delta_r = np.abs(vectors[:, 0])
    elif axis == "y":
        delta_r = np.abs(vectors[:, 1])
    elif axis == "z":
        delta_r = np.abs(vectors[:, 2])
    else:
        raise ValueError('Parameter axis has to be ether "all", "x", "y", or "z"!')
    hist = np.histogram(delta_r, bins, range=(bins[0], bins[-1]))[0]
    hist = hist / (bins[1:] - bins[:-1])
    return hist / len(start_frame)


def van_hove_distinct(
    start_frame: CoordinateFrame,
    end_frame: CoordinateFrame,
    bins: ArrayLike,
    box: ArrayLike = None,
    use_dask: bool = True,
    comp: bool = False,
) -> np.ndarray:
    r"""
    Compute the distinct part of the Van Hove autocorrelation function.

    ..math::
      G(r, t) = \sum_{i, j} \delta(|\vec r_i(0) - \vec r_j(t)| - r)
    """
    if box is None:
        box = start_frame.box.diagonal()
    dimension = len(box)
    N = len(start_frame)
    if use_dask:
        start_frame = darray.from_array(start_frame, chunks=(500, dimension)).reshape(
            1, N, dimension
        )
        end_frame = darray.from_array(end_frame, chunks=(500, dimension)).reshape(
            N, 1, dimension
        )
        dist = (
            (pbc_diff(start_frame, end_frame, box) ** 2).sum(axis=-1) ** 0.5
        ).ravel()
        if np.diff(bins).std() < 1e6:
            dx = bins[0] - bins[1]
            hist = darray.bincount((dist // dx).astype(int), minlength=(len(bins) - 1))
        else:
            hist = darray.histogram(dist, bins=bins)[0]
        return hist.compute() / N
    else:
        if comp:
            dx = bins[1] - bins[0]
            minlength = len(bins) - 1

            def f(x):
                d = (pbc_diff(x, end_frame, box) ** 2).sum(axis=-1) ** 0.5
                return np.bincount((d // dx).astype(int), minlength=minlength)[
                    :minlength
                ]

            hist = sum(f(x) for x in start_frame)
        else:
            dist = (
                pbc_diff(
                    start_frame.reshape(1, -1, 3), end_frame.reshape(-1, 1, 3), box
                )
                ** 2
            ).sum(axis=-1) ** 0.5
            hist = np.histogram(dist, bins=bins)[0]
        return hist / N


def overlap(
    start_frame: CoordinateFrame,
    end_frame: CoordinateFrame,
    radius: float = 0.1,
    mode: str = "self",
) -> float:
    """
    Compute the overlap with a reference configuration defined in a CoordinatesTree.

    Args:
        start_frame: Initial frame, this is only used to get the frame index
        end_frame: The current configuration
        radius: The cutoff radius for the overlap
        mode: Select between "indifferent", "self" or "distict" part of the overlap

    This function is intended to be used with :func:`shifted_correlation`.
    As usual, the first two arguments are used internally, and the remaining ones
    should be defined with :func:`functools.partial`.

    If the overlap of a subset of the system should be calculated, this has to be
    defined through a selection of the reference configurations in the CoordinatesTree.

    Example:
        >>> shifted_correlation(
        ...     partial(overlap, crds_tree=CoordinatesTree(traj), radius=0.11),
        ...     traj
        ... )
    """
    start_PBC, index_PBC = pbc_points(
        start_frame, start_frame.box, index=True, thickness=2 * radius
    )
    start_tree = KDTree(start_PBC)
    dist, index_dist = start_tree.query(end_frame, 1, distance_upper_bound=radius)
    if mode == "indifferent":
        return np.sum(dist <= radius) / len(start_frame)
    index_dist = index_PBC[index_dist]
    index = np.arange(len(start_frame))
    index = index[dist <= radius]
    index_dist = index_dist[dist <= radius]
    if mode == "self":
        return np.sum(index == index_dist) / len(start_frame)
    elif mode == "distinct":
        return np.sum(index != index_dist) / len(start_frame)


def coherent_scattering_function(
    start_frame: CoordinateFrame, end_frame: CoordinateFrame, q: float
) -> np.ndarray:
    """
    Calculate the coherent scattering function.
    """
    box = start_frame.box.diagonal()
    dimension = len(box)

    def scfunc(x, y):
        sqdist = 0
        for i in range(dimension):
            d = x[i] - y[i]
            if d > box[i] / 2:
                d -= box[i]
            if d < -box[i] / 2:
                d += box[i]
            sqdist += d**2
        x = sqdist**0.5 * q
        if x == 0:
            return 1.0
        else:
            return np.sin(x) / x

    return coherent_sum(scfunc, start_frame.pbc, end_frame.pbc) / len(start_frame)


def non_gaussian_parameter(
    start_frame: CoordinateFrame,
    end_frame: CoordinateFrame,
    trajectory: Coordinates = None,
    axis: str = "all",
) -> float:
    """
    Calculate the non-Gaussian parameter.
    ..math:
      \alpha_2 (t) =
        \frac{3}{5}\frac{\langle r_i^4(t)\rangle}{\langle r_i^2(t)\rangle^2} - 1
    """
    if trajectory is None:
        vectors = start_frame - end_frame
    else:
        vectors = displacements_without_drift(start_frame, end_frame, trajectory)
    if axis == "all":
        r = (vectors**2).sum(axis=1)
        dimensions = 3
    elif axis == "xy" or axis == "yx":
        r = (vectors[:, [0, 1]] ** 2).sum(axis=1)
        dimensions = 2
    elif axis == "xz" or axis == "zx":
        r = (vectors[:, [0, 2]] ** 2).sum(axis=1)
        dimensions = 2
    elif axis == "yz" or axis == "zy":
        r = (vectors[:, [1, 2]] ** 2).sum(axis=1)
        dimensions = 2
    elif axis == "x":
        r = vectors[:, 0] ** 2
        dimensions = 1
    elif axis == "y":
        r = vectors[:, 1] ** 2
        dimensions = 1
    elif axis == "z":
        r = vectors[:, 2] ** 2
        dimensions = 1
    else:
        raise ValueError('Parameter axis has to be ether "all", "x", "y", or "z"!')

    return (np.mean(r**2) / ((1 + 2 / dimensions) * (np.mean(r) ** 2))) - 1
